@@ -1,22 +1,3 @@
-# -*- coding: utf-8; Mode: Python -*-
-#
-# Copyright (C) 2011 Lincoln de Sousa <lincoln@comum.org>
-# Copyright 2012 Jeffrey Finkelstein <jeffrey.finkelstein@gmail.com>
-#
-# This file is part of Flask-Restless.
-#
-# Flask-Restless is free software: you can redistribute it and/or modify it
-# under the terms of the GNU Affero General Public License as published by the
-# Free Software Foundation, either version 3 of the License, or (at your
-# option) any later version.
-#
-# Flask-Restless is distributed in the hope that it will be useful, but WITHOUT
-# ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS
-# FOR A PARTICULAR PURPOSE. See the GNU Affero General Public License for more
-# details.
-#
-# You should have received a copy of the GNU Affero General Public License
-# along with Flask-Restless. If not, see <http://www.gnu.org/licenses/>.
 """
     flask.ext.restless.manager
     ~~~~~~~~~~~~~~~~~~~~~~~~~~
@@ -27,7 +8,7 @@
 
     :copyright:2011 by Lincoln de Sousa <lincoln@comum.org>
     :copyright:2012 Jeffrey Finkelstein <jeffrey.finkelstein@gmail.com>
-    :license: GNU AGPLv3, see COPYING for more details
+    :license: GNU AGPLv3+ or BSD
 
 """
 
@@ -76,18 +57,36 @@ class APIManager(object):
     #:    has been registered.
     BLUEPRINTNAME_FORMAT = '%s%s'
 
-    def __init__(self, app=None):
-        """Stores the specified :class:`flask.Flask` application object so that
-        this class can register blueprints on it later.
+    def __init__(self, app=None, flask_sqlalchemy_db=None):
+        """Stores the specified :class:`flask.Flask` application object on
+        which API endpoints will be registered and the
+        :class:`flask.ext.sqlalchemy.SQLAlchemy` object which contains the
+        models which will be exposed.
 
-        If `app` is ``None``, the user must call the :meth:`init_app` method
-        before calling the :meth:`create_api` method.
+        If either `app` or `flask_sqlalchemy_db` is ``None``, the user must
+        call the :meth:`init_app` method before calling the :meth:`create_api`
+        method.
 
         `app` is the :class:`flask.Flask` object containing the user's Flask
         application.
 
+        `flask_sqlalchemy_db` is the :class:`flask.ext.sqlalchemy.SQLAlchemy`
+        object with which `app` has been registered and which contains the
+        database models for which API endpoints will be created.
+
+        For example::
+
+            import flask
+            import flask.ext.restless
+            import flask.ext.sqlalchemy
+
+            app = flask.Flask(__name__)
+            db = flask.ext.sqlalchemy.SQLALchemy(app)
+            apimanager = flask.ext.restless.APIManager(app, db)
+
         """
         self.app = app
+        self.db = flask_sqlalchemy_db
 
     def _next_blueprint_name(self, basename):
         """Returns the next name for a blueprint with the specified base name.
@@ -116,28 +115,37 @@ class APIManager(object):
             next_number = max(existing_numbers) + 1
         return APIManager.BLUEPRINTNAME_FORMAT % (basename, next_number)
 
-    def init_app(self, app):
+    def init_app(self, app, flask_sqlalchemy_db):
         """Stores the specified :class:`flask.Flask` application object on
-        which API endpoints will be registered.
+        which API endpoints will be registered and the
+        :class:`flask.ext.sqlalchemy.SQLAlchemy` object which contains the
+        models which will be exposed.
 
         This is for use in the situation in which this class must be
         instantiated before the :class:`~flask.Flask` application has been
         created. For example::
 
-            apimanager = APIManager()
+            import flask
+            import flask.ext.restless
+            import flask.ext.sqlalchemy
+
+            apimanager = flask.ext.restless.APIManager()
 
             # later...
 
-            app = Flask(__name__)
-            apimanager.init_app(app)
+            app = flask.Flask(__name__)
+            db = flask.ext.sqlalchemy.SQLALchemy(app)
+            apimanager.init_app(app, db)
 
         """
         self.app = app
+        self.db = flask_sqlalchemy_db
 
     def create_api(self, model, methods=READONLY_METHODS, url_prefix='/api',
                    collection_name=None, allow_patch_many=False,
                    allow_functions=False, authentication_required_for=None,
-                   authentication_function=None):
+                   authentication_function=None, include_columns=None,
+                   validation_exceptions=None):
         """Creates a ReSTful API interface as a blueprint and registers it on
         the :class:`flask.Flask` application specified in the constructor to
         this class.
@@ -158,13 +166,12 @@ class APIManager(object):
         object specified in the constructor of this class, so you do *not* need
         to register it yourself.
 
-        ``model`` is the :class:`flask.ext.restless.Entity` class for which a
+        `model` is the :class:`flask.ext.restless.Entity` class for which a
         ReSTful interface will be created. Note this must be a class, not an
         instance of a class.
 
-        ``methods`` specify the HTTP methods which will be made available on
-        the ReSTful API for the specified model, subject to the following
-        caveats:
+        `methods` specify the HTTP methods which will be made available on the
+        ReSTful API for the specified model, subject to the following caveats:
 
         * If :http:method:`get` is in this list, the API will allow getting a
           single instance of the model, getting all instances of the model, and
@@ -185,8 +192,7 @@ class APIManager(object):
         model class to be used in the URL for the ReSTful API created. If this
         is not specified, the lowercase name of the model will be used.
 
-        ``url_prefix`` specifies the URL prefix at which this API will be
-        accessible.
+        `url_prefix` the URL prefix at which this API will be accessible.
 
         If `allow_patch_many` is ``True``, then requests to
         :http:patch:`/api/<collection_name>?q=<searchjson>` will attempt to
@@ -195,6 +201,11 @@ class APIManager(object):
         information on the search query parameter ``q``, see
         :ref:`searchformat`.
 
+        `validation_exceptions` is the tuple of possible exceptions raised by
+        validation of your database models. If this is specified, validation
+        errors will be captured and forwarded to the client in JSON format. For
+        more information on how to use validation, see :ref:`validation`.
+
         If `allow_functions` is ``True``, then requests to
         :http:get:`/api/eval/<collection_name>` will return the result of
         evaluating SQL functions specified in the body of the request. For
@@ -213,6 +224,18 @@ class APIManager(object):
         returns ``True`` if and only if a client is authorized to make a
         request on an endpoint.
 
+        `include_columns` is a list of strings which name the columns of
+        `model` which will be included in the JSON representation of that model
+        provided in response to :http:method:`get` requests. Only the named
+        columns will be included. If this list includes a string which does not
+        name a column in `model`, it will be ignored.
+
+        .. versionadded:: 0.5
+           Added the `include_columns` keyword argument.
+
+        .. versionadded:: 0.5
+           Added the `validation_exceptions` keyword argument.
+
         .. versionadded:: 0.4
            Added the `authentication_required_for` keyword argument.
 
@@ -237,8 +260,9 @@ class APIManager(object):
                    ' authentication_function.')
             raise IllegalArgumentError(msg)
         if collection_name is None:
-            collection_name = model.__name__.lower()
-        methods = frozenset(methods)
+            collection_name = model.__tablename__
+        # convert all method names to upper case
+        methods = frozenset((m.upper() for m in methods))
         # sets of methods used for different types of endpoints
         no_instance_methods = methods & frozenset(('POST', ))
         if allow_patch_many:
@@ -254,8 +278,10 @@ class APIManager(object):
         # the name of the API, for use in creating the view and the blueprint
         apiname = APIManager.APINAME_FORMAT % collection_name
         # the view function for the API for this model
-        api_view = API.as_view(apiname, model, authentication_required_for,
-                               authentication_function)
+        api_view = API.as_view(apiname, self.db.session, model,
+                               authentication_required_for,
+                               authentication_function, include_columns,
+                               validation_exceptions)
         # suffix an integer to apiname according to already existing blueprints
         blueprintname = self._next_blueprint_name(apiname)
         # add the URL rules to the blueprint: the first is for methods on the
@@ -277,7 +303,8 @@ class APIManager(object):
         # evaluating functions on all instances of the specified model
         if allow_functions:
             eval_api_name = apiname + 'eval'
-            eval_api_view = FunctionAPI.as_view(eval_api_name, model)
+            eval_api_view = FunctionAPI.as_view(eval_api_name, self.db.session,
+                                                model)
             eval_endpoint = '/eval' + collection_endpoint
             blueprint.add_url_rule(eval_endpoint, methods=['GET'],
                                    view_func=eval_api_view)
