@@ -1,79 +1,64 @@
-# -*- coding: utf-8; Mode: Python -*-
-#
-# Copyright (C) 2011 Lincoln de Sousa <lincoln@comum.org>
-# Copyright 2012 Jeffrey Finkelstein <jeffrey.finkelstein@gmail.com>
-#
-# This file is part of Flask-Restless.
-#
-# Flask-Restless is free software: you can redistribute it and/or modify it
-# under the terms of the GNU Affero General Public License as published by the
-# Free Software Foundation, either version 3 of the License, or (at your
-# option) any later version.
-#
-# Flask-Restless is distributed in the hope that it will be useful, but WITHOUT
-# ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS
-# FOR A PARTICULAR PURPOSE. See the GNU Affero General Public License for more
-# details.
-#
-# You should have received a copy of the GNU Affero General Public License
-# along with Flask-Restless. If not, see <http://www.gnu.org/licenses/>.
 """
     flask.ext.restless.search
     ~~~~~~~~~~~~~~~~~~~~~~~~~
 
-    Provides querying, searching, and function evaluation on Elixir models.
+    Provides querying, searching, and function evaluation on SQLAlchemy models.
 
     The most important functions in this module are the :func:`create_query`
     and :func:`search` functions, which create a SQLAlchemy query object and
     execute that query on a given model, respectively.
 
-    :copyright:2011 by Lincoln de Sousa <lincoln@comum.org>
-    :copyright:2012 Jeffrey Finkelstein <jeffrey.finkelstein@gmail.com>
-    :license: GNU AGPLv3, see COPYING for more details
+    :copyright: 2011 by Lincoln de Sousa <lincoln@comum.org>
+    :copyright: 2012 Jeffrey Finkelstein <jeffrey.finkelstein@gmail.com>
+    :license: GNU AGPLv3+ or BSD
 
 """
+import inspect
 
 #: The mapping from operator name (as accepted by the search method) to a
 #: function which returns the SQLAlchemy expression corresponding to that
 #: operator.
 #:
-#: The function in each of the values takes three arguments. The first argument
-#: is the field object on which to apply the operator. The second argument is
-#: the second argument to the operator, should one exist. The third argument is
-#: the name of the field. All functions use the first argument, some use the
-#:  second, and few use the third.
+#: Each of these functions accepts either one, two, or three arguments. The
+#: first argument is the field object on which to apply the operator. The
+#: second argument, where it exists, is the second argument to the operator.
+#: The third argument, where it exists, is the name of the field.
 #:
 #: Some operations have multiple names. For example, the equality operation can
 #: be described by the strings ``'=='``, ``'eq'``, ``'equals'``, etc.
 OPERATORS = {
-    '==': lambda f, a, fn: f == a,
-    'eq': lambda f, a, fn: f == a,
-    'equals': lambda f, a, fn: f == a,
-    'equal_to': lambda f, a, fn: f == a,
-    '!=': lambda f, a, fn: f != a,
-    'ne': lambda f, a, fn: f != a,
-    'neq': lambda f, a, fn: f != a,
-    'not_equal_to': lambda f, a, fn: f != a,
-    'does_not_equal': lambda f, a, fn: f != a,
-    '>': lambda f, a, fn: f > a,
-    'gt': lambda f, a, fn: f > a,
-    '<': lambda f, a, fn: f < a,
-    'lt': lambda f, a, fn: f < a,
-    '>=': lambda f, a, fn: f >= a,
-    'ge': lambda f, a, fn: f >= a,
-    'gte': lambda f, a, fn: f >= a,
-    'geq': lambda f, a, fn: f >= a,
-    '<=': lambda f, a, fn: f <= a,
-    'le': lambda f, a, fn: f <= a,
-    'lte': lambda f, a, fn: f <= a,
-    'leq': lambda f, a, fn: f <= a,
-    'like': lambda f, a, fn: f.like(a),
-    'in': lambda f, a, fn: f.in_(a),
-    'not_in': lambda f, a, fn: ~f.in_(a),
-    'is_null': lambda f, a, fn: f == None,
-    'is_not_null': lambda f, a, fn: f != None,
-    'desc': lambda f, a, fn: f.desc,
-    'asc': lambda f, a, fn: f.asc,
+    # Operators which accept a single argument.
+    'is_null': lambda f: f == None,
+    'is_not_null': lambda f: f != None,
+    # TODO what are these?
+    'desc': lambda f: f.desc,
+    'asc': lambda f: f.asc,
+    # Operators which accept two arguments.
+    '==': lambda f, a: f == a,
+    'eq': lambda f, a: f == a,
+    'equals': lambda f, a: f == a,
+    'equal_to': lambda f, a: f == a,
+    '!=': lambda f, a: f != a,
+    'ne': lambda f, a: f != a,
+    'neq': lambda f, a: f != a,
+    'not_equal_to': lambda f, a: f != a,
+    'does_not_equal': lambda f, a: f != a,
+    '>': lambda f, a: f > a,
+    'gt': lambda f, a: f > a,
+    '<': lambda f, a: f < a,
+    'lt': lambda f, a: f < a,
+    '>=': lambda f, a: f >= a,
+    'ge': lambda f, a: f >= a,
+    'gte': lambda f, a: f >= a,
+    'geq': lambda f, a: f >= a,
+    '<=': lambda f, a: f <= a,
+    'le': lambda f, a: f <= a,
+    'lte': lambda f, a: f <= a,
+    'leq': lambda f, a: f <= a,
+    'like': lambda f, a: f.like(a),
+    'in': lambda f, a: f.in_(a),
+    'not_in': lambda f, a: ~f.in_(a),
+    # Operators which accept three arguments.
     # HACK For Python 2.5, unicode dictionary keys are not allowed.
     'has': lambda f, a, fn: f.has(**{str(fn): a}),
     'any': lambda f, a, fn: f.any(**{str(fn): a})
@@ -88,14 +73,6 @@ def _unicode_keys_to_strings(dictionary):
     return dict((str(k), v) for k, v in dictionary.iteritems())
 
 
-class IllegalArgumentError(Exception):
-    """This exception is raised when illegal arguments are provided as input to
-    a function.
-
-    """
-    pass
-
-
 class OrderBy(object):
     """Represents an "order by" in a SQL query expression."""
 
@@ -136,14 +113,17 @@ class Filter(object):
 
         If `argument` is specified, it is the value to place on the right side
         of the operator. If `otherfield` is specified, that field on the model
-        will be placed on the right side of the operator. Exactly one of
-        `arguemnt` and `otherfield` must be specified. If not, a
-        :exc:`IllegalArgumentError` will be raised.
+        will be placed on the right side of the operator.
+
+        .. admonition:: About `argument` and `otherfield`
+
+           Some operators don't need either argument and some need exactly one.
+           However, this constructor will not raise any errors or otherwise
+           inform you of which situation you are in; it is basically just a
+           named tuple. Calling code must handle errors caused by missing
+           required arguments.
 
         """
-        if (argument and otherfield) or not (argument or otherfield):
-            raise IllegalArgumentError('Must specify exactly one of argument'
-                                       ' and otherfield')
         self.fieldname = fieldname
         self.operator = operator
         self.argument = argument
@@ -174,10 +154,6 @@ class Filter(object):
         name of the other field of the model to which the operator will be
         applied.
 
-        Raises a :exc:`IllegalArgumentError` if both ``'val'`` and ``'field'``
-        are in `dictionary` or if neither ``'val'`` not ``'field'`` are in
-        `dictionary`.
-
         """
         fieldname = dictionary.get('name')
         operator = dictionary.get('op')
@@ -244,13 +220,9 @@ class SearchParameters(object):
         The provided dictionary may have other key/value pairs, but they are
         ignored.
 
-        Raises a :exc:`IllegalArgumentError` if there is an error instantiating
-        one of the :class:`Filter` objects.
-
         """
         # for the sake of brevity...
         from_dict = Filter.from_dictionary
-        # may raise IllegalArgumentError
         filters = [from_dict(f) for f in dictionary.get('filters', [])]
         # HACK In Python 2.5, unicode dictionary keys are not allowed.
         order_by_list = dictionary.get('order_by', [])
@@ -282,40 +254,66 @@ class QueryBuilder(object):
         are given by the keys of :data:`OPERATORS`. For more information on
         recognized search operators, see :ref:`search`.
 
-        If ``relation`` is not ``None``, the returned search parameter will
-        correspond to a search on the field named ``fieldname`` on the entity
-        related to ``model`` whose name, as a string, is ``relation``.
+        If `relation` is not ``None``, the returned search parameter will
+        correspond to a search on the field named `fieldname` on the entity
+        related to `model` whose name, as a string, is `relation`.
 
-        ``model`` is an instance of a :class:`elixir.entity.Entity` being
+        `model` is an instance of a SQLAlchemy declarative model being
         searched.
 
-        ``fieldname`` is the name of the field of ``model`` to which the
-        operation will be applied as part of the search. If ``relation`` is
-        specified, the operation will be applied to the field with name
-        ``fieldname`` on the entity related to ``model`` whose name, as a
-        string, is ``relation``.
+        `fieldname` is the name of the field of `model` to which the operation
+        will be applied as part of the search. If `relation` is specified, the
+        operation will be applied to the field with name `fieldname` on the
+        entity related to `model` whose name, as a string, is `relation`.
 
-        ``operation`` is a string representating the operation which will be
+        `operation` is a string representating the operation which will be
          executed between the field and the argument received. For example,
          ``'gt'``, ``'lt'``, ``'like'``, ``'in'`` etc.
 
-        ``argument`` is the argument to which to apply the ``operator``.
+        `argument` is the argument to which to apply the `operator`.
 
-        ``relation`` is the name of the relationship attribute of ``model`` to
+        `relation` is the name of the relationship attribute of `model` to
         which the operation will be applied as part of the search, or ``None``
         if this function should not use a related entity in the search.
 
+        This function raises the following errors:
+        * :exc:`KeyError` if the `operator` is unknown (that is, not in
+          :data:`OPERATORS`)
+        * :exc:`TypeError` if an incorrect number of arguments are provided for
+          the operation (for example, if `operation` is `'=='` but no
+          `argument` is provided)
+        * :exc:`AttributeError` if no column with name `fieldname` or
+          `relation` exists on `model`
+
         """
+        # raises KeyError if operator not in OPERATORS
+        opfunc = OPERATORS[operator]
+        argspec = inspect.getargspec(opfunc)
+        # in Python 2.6 or later, this should be `argspec.args`
+        numargs = len(argspec[0])
+        # raises AttributeError if `fieldname` or `relation` does not exist
         field = getattr(model, relation or fieldname)
-        return OPERATORS.get(operator)(field, argument, fieldname)
+        # each of these will raise a TypeError if the wrong number of argments
+        # is supplied to `opfunc`.
+        if numargs == 1:
+            return opfunc(field)
+        if argument is None:
+            raise TypeError
+        if numargs == 2:
+            return opfunc(field, argument)
+        return opfunc(field, argument, fieldname)
 
     @staticmethod
     def _create_filters(model, search_params):
-        """Returns the list of operations on ``model`` specified in the
-        :attr:`filters` attribute on the ``search_params`` object.
+        """Returns the list of operations on `model` specified in the
+        :attr:`filters` attribute on the `search_params` object.
 
-        ``search_params`` is an instance of the :class:`SearchParameters`
-        class whose fields represent the parameters of the search.
+        `search-params` is an instance of the :class:`SearchParameters` class
+        whose fields represent the parameters of the search.
+
+        Raises one of :exc:`AttributeError`, :exc:`KeyError`, or
+        :exc:`TypeError` if there is a problem creating the query. See the
+        documentation for :func:`_create_operation` for more information.
 
         """
         filters = []
@@ -329,23 +327,23 @@ class QueryBuilder(object):
             # get the other field to which to compare, if it exists
             if filt.otherfield:
                 val = getattr(model, filt.otherfield)
-            param = QueryBuilder._create_operation(model, fname, filt.operator,
-                                                   val, relation)
+            # for the sake of brevity...
+            create_op = QueryBuilder._create_operation
+            param = create_op(model, fname, filt.operator, val, relation)
             filters.append(param)
         return filters
 
     @staticmethod
-    def create_query(model, search_params):
+    def create_query(session, model, search_params):
         """Builds an SQLAlchemy query instance based on the search parameters
         present in ``search_params``, an instance of :class:`SearchParameters`.
 
         This method returns a SQLAlchemy query in which all matched instances
         meet the requirements specified in ``search_params``.
 
-        ``model`` is an :class:`elixir.entity.Entity` on which to create a
-        query.
+        `model` is SQLAlchemy declarative model on which to create a query.
 
-        ``search_params`` is an instance of :class:`SearchParameters` which
+        `search_params` is an instance of :class:`SearchParameters` which
         specify the filters, order, limit, offset, etc. of the query.
 
         Building the query proceeds in this order:
@@ -354,9 +352,14 @@ class QueryBuilder(object):
         3. limiting the query
         4. offsetting the query
 
+        Raises one of :exc:`AttributeError`, :exc:`KeyError`, or
+        :exc:`TypeError` if there is a problem creating the query. See the
+        documentation for :func:`_create_operation` for more information.
+
         """
         # Adding field filters
-        query = model.query
+        query = session.query(model)
+        # may raise exception here
         filters = QueryBuilder._create_filters(model, search_params)
         for filt in filters:
             query = query.filter(filt)
@@ -375,18 +378,18 @@ class QueryBuilder(object):
         return query
 
 
-def create_query(model, searchparams):
-    """Returns a SQLAlchemy query object on the given ``model`` where the
-    search for the query is defined by ``searchparams``.
+def create_query(session, model, searchparams):
+    """Returns a SQLAlchemy query object on the given `model` where the search
+    for the query is defined by `searchparams`.
 
-    The returned query matches the set of all instances of ``model`` which meet
-    the parameters of the search given by ``searchparams``. For more
-    information on search parameters, see :ref:`search`.
+    The returned query matches the set of all instances of `model` which meet
+    the parameters of the search given by `searchparams`. For more information
+    on search parameters, see :ref:`search`.
 
-    ``model`` is a :class:`elixir.Entity` representing the database model to
-    query.
+    `model` is a SQLAlchemy declarative model representing the database model
+    to query.
 
-    ``searchparams`` is either a dictionary (as parsed from a JSON request from
+    `searchparams` is either a dictionary (as parsed from a JSON request from
     the client, for example) or a :class:`SearchParameters` instance defining
     the parameters of the query (as returned by
     :func:`SearchParameters.from_dictionary`, for example).
@@ -394,10 +397,10 @@ def create_query(model, searchparams):
     """
     if isinstance(searchparams, dict):
         searchparams = SearchParameters.from_dictionary(searchparams)
-    return QueryBuilder.create_query(model, searchparams)
+    return QueryBuilder.create_query(session, model, searchparams)
 
 
-def search(model, search_params):
+def search(session, model, search_params):
     """Performs the search specified by the given parameters on the model
     specified in the constructor of this class.
 
@@ -413,10 +416,10 @@ def search(model, search_params):
     results are found and :exc:`sqlalchemy.orm.exc.MultipleResultsFound` if
     multiple results are found.
 
-    ``model`` is a :class:`elixir.Entity` representing the database model to
-    query.
+    `model` is a SQLAlchemy declarative model class representing the database
+    model to query.
 
-    ``search_params`` is a dictionary containing all available search
+    `search_params` is a dictionary containing all available search
     parameters. For more information on available search parameters, see
     :ref:`search`. Implementation note: this dictionary will be converted to a
     :class:`SearchParameters` object when the :func:`create_query` function is
@@ -427,7 +430,7 @@ def search(model, search_params):
     # corresponding value is anything except those values which evaluate to
     # False (False, 0, the empty string, the empty list, etc.).
     is_single = search_params.get('single')
-    query = create_query(model, search_params)
+    query = create_query(session, model, search_params)
     if is_single:
         # may raise NoResultFound or MultipleResultsFound
         return query.one()
